#include "stdafx.h"
#include "ScanFinger.h"
#include "ScanFingerDoc.h"
#include "ScanFingerView.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#pragma warning(suppress: 26409 26433 26440 26477)
IMPLEMENT_DYNCREATE(CScanFingerView, CScrollView)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CScanFingerView, CScrollView) //NOLINT(modernize-avoid-c-arrays)
  ON_COMMAND(ID_EDIT_COPY, &CScanFingerView::OnEditCopy)
  ON_UPDATE_COMMAND_UI(ID_EDIT_COPY, &CScanFingerView::OnUpdateEditCopy)
  ON_WM_ERASEBKGND()
  ON_WM_CONTEXTMENU()
  ON_COMMAND(ID_ZOOM_ZOOMIN, &CScanFingerView::OnZoomZoomin)
  ON_COMMAND(ID_ZOOM_ZOOMOUT, &CScanFingerView::OnZoomZoomout)
  ON_UPDATE_COMMAND_UI(ID_ZOOM_ZOOMIN, &CScanFingerView::OnUpdateZoomZoomin)
  ON_UPDATE_COMMAND_UI(ID_ZOOM_ZOOMOUT, &CScanFingerView::OnUpdateZoomZoomout)
  ON_COMMAND(ID_ZOOM_ACTUAL, &CScanFingerView::OnZoomActual)
  ON_UPDATE_COMMAND_UI(ID_ZOOM_ACTUAL, &CScanFingerView::OnUpdateZoomActual)
  ON_COMMAND(ID_ZOOM_BESTFIT, &CScanFingerView::OnZoomBestfit)
  ON_UPDATE_COMMAND_UI(ID_ZOOM_BESTFIT, &CScanFingerView::OnUpdateZoomBestfit)
  ON_WM_SIZE()
END_MESSAGE_MAP()

CScanFingerView::CScanFingerView() noexcept : m_nCurrentZoomLevel{100},
                                              m_bSizeToFit{true}
{
}

#pragma warning(suppress: 26429)
void CScanFingerView::OnDraw(CDC* pDC)
{
  switch (theApp.m_ScanState)
  {
    case CScanFingerApp::State::ErrorOpening:
    {
      //Get the client rect
      CRect rClient;
      GetClientRect(&rClient);

      //Prompt the user to scan their finger with some help text
      CGdiObject* pOldFont{pDC->SelectStockObject(DEFAULT_GUI_FONT)};
      pDC->FillSolidRect(&rClient, RGB(255, 255, 255));
      pDC->DrawText(theApp.m_sScanImageMessage, &rClient, DT_WORDBREAK | DT_CENTER | DT_VCENTER);
#pragma warning(suppress: 26486)
      pDC->SelectObject(pOldFont);
      break;
    }
    case CScanFingerApp::State::Scanning: //deliberate fallthrough
    case CScanFingerApp::State::ScanAcquired:
    {
      //Get the client rect
      CRect rClient;
      GetClientRect(&rClient);
      CBrush br;
      br.CreateHatchBrush(HS_DIAGCROSS, RGB(255, 255, 255));
      FillOutsideRect(pDC, &br);

      //Draw the image if the document has one
      if (theApp.m_pScanImage != nullptr)
      {
        //Determine where the image should be drawn
        const CSize drawSize{DrawSize(rClient)};
        const CRect rDraw{0, 0, drawSize.cx, drawSize.cy};

        //Create the GDI+ Graphics object
        Gdiplus::Graphics graphics{pDC->GetSafeHdc()};
        if (graphics.GetLastStatus() == Gdiplus::Ok)
        {
          //Setup appropriate values for the graphics object
          graphics.SetSmoothingMode(Gdiplus::SmoothingModeHighQuality);

          //Render the image
          graphics.DrawImage(theApp.m_pScanImage.get(), Gdiplus::RectF(0, 0, static_cast<Gdiplus::REAL>(rDraw.Width()), static_cast<Gdiplus::REAL>(rDraw.Height())));
        }
      }
      else
      {
        //Prompt the user to scan their finger with some help text
        CGdiObject* pOldFont{pDC->SelectStockObject(DEFAULT_GUI_FONT)};
        pDC->FillSolidRect(&rClient, RGB(255, 255, 255));
        pDC->DrawText(theApp.m_sScanImageMessage, &rClient, DT_WORDBREAK | DT_CENTER | DT_VCENTER);
#pragma warning(suppress: 26486)
        pDC->SelectObject(pOldFont);
      }
      break;
    }
    default:
    {
      break;
    }
  }
}

#ifdef _DEBUG
void CScanFingerView::AssertValid() const
{
  __super::AssertValid();
}

void CScanFingerView::Dump(CDumpContext& dc) const
{
  __super::Dump(dc);
}
#endif //_DEBUG

#pragma warning(suppress: 26434 26440)
CScanFingerDoc* CScanFingerView::GetDocument() const
{
  //Validate our parameters
  ASSERT(m_pDocument != nullptr);
  ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CScanFingerDoc)));

#pragma warning(suppress: 26466)
  return static_cast<CScanFingerDoc*>(m_pDocument);
}

void CScanFingerView::OnInitialUpdate()
{
  //Let the base class do its thing
  __super::OnInitialUpdate();

  //Start with a default sized view
  SetScrollSizes(MM_TEXT, CSize{400, 400});
}

void CScanFingerView::OnEditCopy()
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(theApp.m_pScanImage != nullptr);

  //Copy from the DDB to DIB and transfer to the clipboard
  CBitmap bitmap;
#pragma warning(suppress: 26490)
  if (theApp.m_pScanImage->GetHBITMAP(Gdiplus::Color(0, 0, 0), reinterpret_cast<HBITMAP*>(&bitmap.m_hObject)) != Gdiplus::Ok)
    return;
  HGLOBAL hDib{DDBToDIB(&bitmap, nullptr)};
  if (hDib && OpenClipboard())
  {
    EmptyClipboard();
    SetClipboardData(CF_DIB, hDib);
    CloseClipboard();
    bitmap.Detach();
  }
}

#pragma warning(suppress: 26429)
void CScanFingerView::OnUpdateEditCopy(CCmdUI* pCmdUI)
{
#pragma warning(suppress: 26486)
  pCmdUI->Enable(theApp.m_pScanImage != nullptr);
}

HGLOBAL CScanFingerView::DDBToDIB(_In_ const CBitmap* pBitmap, _In_opt_ CPalette* pPalette)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(pBitmap != nullptr);

  //Get the palette to use
  CPalette localPalette;
  if (pPalette == nullptr)
  {
    localPalette.CreateStockObject(DEFAULT_PALETTE);
    pPalette = &localPalette;
  }

  //Get info about the DDB
  BITMAP bm{};
  pBitmap->GetObject(sizeof(bm), &bm);

  //Initialize the BITMAPINFOHEADER
  BITMAPINFOHEADER bi{};
  bi.biSize = sizeof(BITMAPINFOHEADER);
  bi.biWidth = bm.bmWidth;
  bi.biHeight = bm.bmHeight;
  bi.biPlanes = 1;
#pragma warning(suppress: 26472)
  bi.biBitCount = static_cast<WORD>(bm.bmPlanes * bm.bmBitsPixel);
  bi.biCompression = BI_RGB;

  //Compute the size of the infoheader and the color table
  int nColors{0};
  if (bi.biBitCount <= 8)
    nColors = (1 << bi.biBitCount);
  SIZE_T dwLen{bi.biSize + (nColors * sizeof(RGBQUAD))};

  //We need a device context to get the DDB from
  CDC* pDC{GetDC()};
  if (pDC == nullptr)
  {
    TRACE(_T("CScanFingerView::DDBToDIB, Failed in call to GetDC, Error:%u\n"), GetLastError());
#pragma warning(suppress: 26487)
    return nullptr;
  }
  CPalette* pOldPalette{pDC->SelectPalette(pPalette, FALSE)};
  pDC->RealizePalette();

  //Allocate enough memory to hold BITMAPINFOHEADER and color table
  auto lpBi{static_cast<LPBITMAPINFOHEADER>(GlobalAlloc(GMEM_FIXED, dwLen))};
  if (lpBi == nullptr)
  {
    TRACE(_T("CScanFingerView::DDBToDIB, Failed to allocate memory for the DIB, Error:%u\n"), GetLastError());
    pDC->SelectPalette(pOldPalette, FALSE);
#pragma warning(suppress: 26486)
    ReleaseDC(pDC);
#pragma warning(suppress: 26487)
    return nullptr;
  }

  //Transfer the BITMAPINFOHEADER into the freshly allocated memory handle
#pragma warning(suppress: 6386)
  memcpy_s(lpBi, dwLen, &bi, dwLen);

  //Call GetDibBits with a nullptr lpBits, so the device driver will
  //calculate the biSizeImage field
  auto hBitmap{static_cast<HBITMAP>(pBitmap->GetSafeHandle())};
  if (hBitmap == nullptr)
  {
    TRACE(_T("CScanFingerView::DDBToDIB, Failed to get bitmap handle. Error:%u\n"), GetLastError());
    GlobalFree(lpBi);
    pDC->SelectPalette(pOldPalette, FALSE);
#pragma warning(suppress: 26486)
    ReleaseDC(pDC);
#pragma warning(suppress: 26487)
    return nullptr;
  }
#pragma warning(suppress: 26490)
  GetDIBits(pDC->GetSafeHdc(), hBitmap, 0L, bi.biHeight, nullptr, reinterpret_cast<LPBITMAPINFO>(lpBi), DIB_RGB_COLORS);

  //Transfer back to the stack variable
  memcpy_s(&bi, sizeof(bi), lpBi, sizeof(BITMAPINFOHEADER));

  //If the driver did not fill in the biSizeImage field, then compute it.
  //Each scan line of the image is aligned on a DWORD (32 bit) boundary
  if (bi.biSizeImage == 0)
    bi.biSizeImage = ((((bi.biWidth * bi.biBitCount) + 31) & ~31) / 8) * bi.biHeight;

  //Realloc the buffer so that it can hold all the bits
  dwLen += bi.biSizeImage;
  HGLOBAL hNew{GlobalReAlloc(lpBi, dwLen, GMEM_MOVEABLE)};
  if (hNew == nullptr)
  {
    TRACE(_T("CScanFingerView::DDBToDIB, Failed to realloc the memory for the DIB, Error:%u\n"), GetLastError());
#pragma warning(suppress: 6001)
    GlobalFree(lpBi);
    pDC->SelectPalette(pOldPalette, FALSE);
#pragma warning(suppress: 26486)
    ReleaseDC(pDC);
#pragma warning(suppress: 26487)
    return nullptr;
  }

  //Lock down the memory
  lpBi = static_cast<LPBITMAPINFOHEADER>(GlobalLock(hNew));
  if (lpBi == nullptr)
  {
    TRACE(_T("CScanFingerView::DDBToDIB, Failed to lock the memory for the DIB, Error:%u\n"), GetLastError());
    GlobalFree(hNew);
    pDC->SelectPalette(pOldPalette, FALSE);
#pragma warning(suppress: 26486)
    ReleaseDC(pDC);
    return nullptr;
  }

  //Finally get the DIB
#pragma warning(suppress: 26481 26490)
  LPBYTE lpBits{reinterpret_cast<LPBYTE>(lpBi) + (bi.biSize + (nColors * sizeof(RGBQUAD)))};
#pragma warning(suppress: 26490)
  if (!GetDIBits(pDC->GetSafeHdc(), hBitmap, 0, bi.biHeight, lpBits, reinterpret_cast<LPBITMAPINFO>(lpBi), DIB_RGB_COLORS))
  {
    TRACE(_T("CScanFingerView::DDBToDIB, Failed to get the DDB's bits, Error:%u\n"), GetLastError());
    GlobalFree(hNew);
    pDC->SelectPalette(pOldPalette, FALSE);
#pragma warning(suppress: 26486)
    ReleaseDC(pDC);
#pragma warning(suppress: 26487)
    return nullptr;
  }

  //Tidy up the DC before we return
  pDC->SelectPalette(pOldPalette, FALSE);
#pragma warning(suppress: 26486)
  ReleaseDC(pDC);

  return hNew;
}

CSize CScanFingerView::DrawSize(_In_ const CRect& rClient)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(theApp.m_pScanImage != nullptr);

  const UINT nImageWidth{theApp.m_pScanImage->GetWidth()};
  const UINT nImageHeight{theApp.m_pScanImage->GetHeight()};
  if (m_bSizeToFit)
  {
    const UINT nClientWidth = rClient.Width();
    const UINT nClientHeight = rClient.Height();

    if ((nImageWidth < nClientWidth) && (nImageHeight < nClientHeight))
    {
      m_nCurrentZoomLevel = nClientWidth * 100 / nImageWidth;
      const UINT nZoom{nClientHeight * 100 / nImageHeight};
      if (nZoom < m_nCurrentZoomLevel)
        m_nCurrentZoomLevel = nZoom;
    }
    else if ((nImageWidth > nClientWidth) || (nImageHeight > nClientHeight))
    {
      m_nCurrentZoomLevel = nClientWidth * 100 / nImageWidth;
      const UINT nZoom{nClientHeight * 100 / nImageHeight};
      if (nZoom < m_nCurrentZoomLevel)
        m_nCurrentZoomLevel = nZoom;
    }
  }

#pragma warning(suppress: 26472)
  return {static_cast<int>(nImageWidth * m_nCurrentZoomLevel / 100), static_cast<int>(nImageHeight * m_nCurrentZoomLevel / 100)};
}

void CScanFingerView::OnUpdate(CView* /*pSender*/, LPARAM /*lHint*/, CObject* /*pHint*/)
{
  //Set the scroll view extend to that of the picture in the document
  if (theApp.m_pScanImage != nullptr)
  {
    CRect r;
    GetClientRect(&r);
    SetScrollSizes(MM_TEXT, DrawSize(r));
  }
  else
    SetScrollSizes(MM_TEXT, CSize(0, 0));
  Invalidate();
  UpdateWindow();
}

#pragma warning(suppress: 26434 26440)
BOOL CScanFingerView::OnEraseBkgnd(CDC* /*pDC*/)
{
  return TRUE;
}

BOOL CScanFingerView::PreTranslateMessage(MSG* pMsg)
{
  if (((((pMsg->message == WM_KEYDOWN) || (pMsg->message == WM_SYSKEYDOWN)) &&
      (pMsg->wParam == VK_F10) && (GetKeyState(VK_SHIFT) & 0x8000)) != 0) ||
      (pMsg->message == WM_CONTEXTMENU))
  {
    CPoint point;
    GetCursorPos(&point);
    OnContextMenu(nullptr, point);
    return TRUE;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_DOWN))
  {
    SendMessage(WM_VSCROLL, SB_LINEDOWN);
    return TRUE;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_UP))
  {
    SendMessage(WM_VSCROLL, SB_LINEUP);
    return TRUE;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_LEFT))
  {
    SendMessage(WM_HSCROLL, SB_LINEUP);
    return TRUE;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_RIGHT))
  {
    SendMessage(WM_HSCROLL, SB_LINEDOWN);
    return TRUE;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_NEXT))
  {
    SendMessage(WM_VSCROLL, SB_PAGEDOWN);
    return TRUE;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_PRIOR))
  {
    SendMessage(WM_VSCROLL, SB_PAGEUP);
    return TRUE;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_HOME))
  {
    SendMessage(WM_HSCROLL, SB_PAGEUP);
    return TRUE;
  }
  else if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_END))
  {
    SendMessage(WM_HSCROLL, SB_PAGEDOWN);
    return TRUE;
  }
  else
  {
    //Let the base class do its thing
    return __super::PreTranslateMessage(pMsg);
  }
}

#pragma warning(suppress: 26434)
void CScanFingerView::OnContextMenu(CWnd* /*pWnd*/, CPoint point)
{
  CMenu menu;
  if (!menu.LoadMenu(IDR_POPUP_VIEW))
    return;

  CMenu* pPopup{menu.GetSubMenu(0)};
#pragma warning(suppress: 26496)
  AFXASSUME(pPopup != nullptr);

#pragma warning(suppress: 26429)
  CWnd* pWndPopupOwner{this};
  while (pWndPopupOwner->GetStyle() & WS_CHILD)
    pWndPopupOwner = pWndPopupOwner->GetParent();

#pragma warning(suppress: 26486)
  pPopup->TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, pWndPopupOwner);
}

void CScanFingerView::OnZoomZoomin()
{
  m_nCurrentZoomLevel = (m_nCurrentZoomLevel / 20 * 20) + 20;
  m_bSizeToFit = false;
  CRect r;
  GetClientRect(&r);
  SetScrollSizes(MM_TEXT, DrawSize(r));
  Invalidate();
  UpdateWindow();
}

void CScanFingerView::OnZoomZoomout()
{
  m_nCurrentZoomLevel = (m_nCurrentZoomLevel / 20 * 20) - 20;
  m_bSizeToFit = false;
  CRect r;
  GetClientRect(&r);
  SetScrollSizes(MM_TEXT, DrawSize(r));
  Invalidate();
  UpdateWindow();
}

#pragma warning(suppress: 26429)
void CScanFingerView::OnUpdateZoomZoomin(CCmdUI* pCmdUI)
{
  if (theApp.m_pScanImage != nullptr)
#pragma warning(suppress: 26486)
    pCmdUI->Enable(m_bSizeToFit || (m_nCurrentZoomLevel < 300));
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

#pragma warning(suppress: 26429)
void CScanFingerView::OnUpdateZoomZoomout(CCmdUI* pCmdUI)
{
  if (theApp.m_pScanImage != nullptr)
#pragma warning(suppress: 26486)
    pCmdUI->Enable(m_bSizeToFit || (m_nCurrentZoomLevel > 20));
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

void CScanFingerView::OnZoomActual()
{
  m_nCurrentZoomLevel = 100;
  m_bSizeToFit = false;
  CRect r;
  GetClientRect(&r);
  SetScrollSizes(MM_TEXT, DrawSize(r));
  Invalidate();
  UpdateWindow();
}

#pragma warning(suppress: 26429)
void CScanFingerView::OnUpdateZoomActual(CCmdUI* pCmdUI)
{
  if (theApp.m_pScanImage != nullptr)
#pragma warning(suppress: 26486)
    pCmdUI->Enable(m_bSizeToFit || (m_nCurrentZoomLevel != 100));
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

void CScanFingerView::OnZoomBestfit()
{
  m_bSizeToFit = !m_bSizeToFit;
  CRect r;
  GetClientRect(&r);
  SetScrollSizes(MM_TEXT, DrawSize(r));
  Invalidate();
  UpdateWindow();
}

#pragma warning(suppress: 26429)
void CScanFingerView::OnUpdateZoomBestfit(CCmdUI* pCmdUI)
{
  if (theApp.m_pScanImage != nullptr)
#pragma warning(suppress: 26486)
    pCmdUI->SetCheck(m_bSizeToFit);
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

#pragma warning(suppress: 26434)
void CScanFingerView::OnSize(UINT nType, int cx, int cy)
{
  //Update the scroll size if size to fit is turned on
  if (m_bSizeToFit && (theApp.m_pScanImage != nullptr))
  {
    CRect r;
    GetClientRect(&r);
    SetScrollSizes(MM_TEXT, DrawSize(r));
  }

  //Let the base class do its thing
  __super::OnSize(nType, cx, cy);
}
